#------------------------------------------------------------------------------
#
#   Interesting Times - Constants
#
#------------------------------------------------------------------------------

from numpy import array
from vectors import vector

#  Size of a tile cell in pixels
CELL_WIDTH = 16
CELL_HEIGHT = 16
CELL_SIZE = vector(CELL_WIDTH, CELL_HEIGHT)

#  Team numbers
NO_TEAM = 0
BLUE_TEAM = 1
RED_TEAM = 2

#  Timing
#TICKS_PER_MINUTE = 1
START_TIME = (6, 0) # (hours, minutes)
PRINTING_TIME = (3, 0)
HOUSE_UPDATE_TIME = (6, 0)
DELIVERY_START_TIME = (6, 0)
DELIVERY_END_TIME = (9, 0)
PURCHASING_TIME = (9, 0)
REPORTER_PAY_TIME = (17, 0)

#  Initial values
INITIAL_CASH = 500

#  Costs and prices
PRINTING_COST = 1
SUBSCRIPTION_PRICE = 2
NEWSSTAND_PRICE = 3
REPORTER_WAGE = 10
RUNNER_WAGE = 5

#  News item generation and collection
MIN_NEWS_VALUE = 1
MAX_NEWS_VALUE = 10
NEWS_GENERATION_RATE = 0.05 # items per game minute
REPORTER_DETECTION_RANGE = 7 * CELL_WIDTH
REPORTER_PICKUP_RANGE = 2 * CELL_WIDTH

#  Limits
MAX_REPORTERS = 5
MAX_RUNNERS = 5

#  Conversion rate between news value and loyalty increase
LOYALTY_PER_COLUMN_INCH = 0.4

#  Loyalty value for 100% loyalty bar reading
MAX_LOYALTY = 100

#  Factor by which loyalty decreases each day
LOYALTY_DECAY_RATE = 0.7

#  Loyalty below which a house will consider unsubscribing
LOYALTY_UNSUBSCRIPTION_THRESHOLD = 0.3 * MAX_LOYALTY

#  Neighbourly influence above which a house will consider subscribing
LOYALTY_SUBSCRIPTION_THRESHOLD = 0.5 * MAX_LOYALTY

#  Fraction of neighbour's loyalty affecting subscription decisions
NEIGHBOURLY_INFLUENCE_FACTOR = 0.4
NEIGHBOURLY_INFLUENCE_FALLOFF = 0.5

zero = vector(0, 0)
west = vector(-1, 0)
east = vector(1, 0)
north = vector(0, -1)
south = vector(0, 1)
all_directions = array([(0, 1), (1, 0), (0, -1), (-1, 0)])

game_speeds = [0.125, 0.5, 2.0] # game minutes per tick
game_speed_labels = ["Slow", "Medium", "Fast"]

SLOWEST_GAME_SPEED = game_speeds[0]

def other_team(s):
	return 3 - s
