#------------------------------------------------------------------------------
#
#   Interesting Times - Reporter
#
#------------------------------------------------------------------------------

from random import choice
from albow.resource import get_image, get_sound
from vectors import vector, distance
from actor import Actor, IDLE_MODE
from it_collections import CellSet, CellDict
import constants as k

GO_TO_TARGET_MODE = 1
WRITEUP_MODE = 2

class Reporter(Actor):

	images = [
		get_image("reporter-blue.png", border = 2),
		get_image("reporter-red.png", border = 2),
	]
	
	writeup_sound = get_sound("typewriter.ogg")

	foot_offset = vector(19, 30)
	writeup_delay = 3
	distances_to_target = {}
	
	def __init__(self, *args, **kwds):
		Actor.__init__(self, *args, **kwds)
		self.active = True

	def set_goal(self, cell):
		#print "Reporter.set_goal:", cell ###
		self.target_cell = cell
		self.plan_route_to_target()
		self.mode = GO_TO_TARGET_MODE

	def think(self):
		#print "Reporter.think" ###
		self.collect_nearby_news()
		if self.mode == GO_TO_TARGET_MODE:
			if self.reached_target():
				#print "Reporter has reached target" ###
				self.stop()
			else:
				self.head_towards_target()
		#elif self.mode == WRITEUP_MODE:
		#	self.do_writeup_delay()
		else:
			self.seek_news()
	
	def seek_news(self):
		#print "Reporter.seek_news" ###
		limited = not self.team.auto_play
		range = k.REPORTER_DETECTION_RANGE
		here = self.position
		def filter(item):
			if limited:
				return distance(here, item.position) <= range
			else:
				return True
		news = [item for item in self.level.news_items
			if filter(item)]
		if news:
			news.sort(key = lambda item: -item.news_value)
			item = choice(news[:3])
			cell = self.city.point_to_cell(item.position)
			self.set_goal(cell)

	def plan_route_to_target(self):
		city = self.city
		work = CellSet()
		dists = CellDict()
		target = self.target_cell
		dists[target] = 0
		work.add(target)
		while work:
			cell = work.pop()
			dist = dists[cell] + 1
			for adj in city.iter_adjacent_road_cells(cell):
				if adj not in dists:
					dists[adj] = dist
					work.add(adj)
		self.distances_to_target = dists

	def reached_target(self):
		target = self.target_cell
		return target is not None and tuple(target) == tuple(self.cell)

	def head_towards_target(self):
		best_cell = None
		for cell in self.city.iter_adjacent_road_cells(self.cell):
			dist = self.distances_to_target.get(cell)
			if dist is not None and (best_cell is None or dist < best_dist):
				best_dist = dist
				best_cell = cell
		if best_cell is not None:
			self.velocity = best_cell - self.cell
		else:
			self.stop()

	def collect_nearby_news(self):
		level = self.level
		p = self.position
		for item in level.news_items:
			if distance(p, item.position) <= k.REPORTER_PICKUP_RANGE:
				level.remove_news_item(item)
				team = self.team
				if team.team_no == k.BLUE_TEAM:
					self.writeup_sound.play()
				team.collect_news(item.news_value)
				#self.delay = self.writeup_delay
				#self.mode = WRITEUP_MODE
				#print "Reporter: entering writeup mode" ###
	
	def do_writeup_delay(self):
		#print "Reporter.do_writeup_delay" ###
		self.delay -= 1
		if self.delay == 0:
			self.mode = GO_TO_TARGET_MODE

